/*
 * Copyright (c) 1990,1991 Regents of The University of Michigan.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation, and that the name of The University
 * of Michigan not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission. This software is supplied as is without expressed or
 * implied warranties of any kind.
 *
 *	Research Systems Unix Group
 *	The University of Michigan
 *	c/o Mike Clark
 *	535 W. William Street
 *	Ann Arbor, Michigan
 *	+1-313-763-0525
 *	netatalk@itd.umich.edu
 */

#ifndef AFPD_DESKTOP_H
#define AFPD_DESKTOP_H 1

#include <atalk/globals.h>

#include "volume.h"

#define APPLEDESKTOP ".AppleDesktop"

struct savedt {
    uint8_t	sdt_creator[4];
    int		sdt_fd;
    int		sdt_index;
    short	sdt_vid;
};

typedef unsigned char CreatorType[4];

extern char	*dtfile(const struct vol *, uint8_t [], char *);
extern char	*mtoupath(const struct vol *, char *, cnid_t, int utf8);
extern char	*utompath(const struct vol *, char *, cnid_t, int utf8);

extern int setdeskmode(const struct vol *vol, const mode_t mode);
extern int setdeskowner(const struct vol *vol, uid_t uid, gid_t gid);

/* FP functions */
int afp_opendt(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
               size_t *rbuflen);
int afp_addcomment(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                   size_t *rbuflen);
int afp_getcomment(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                   size_t *rbuflen);
int afp_rmvcomment(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                   size_t *rbuflen);
int afp_addappl(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);
int afp_rmvappl(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);
int afp_getappl(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);
int afp_closedt(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);
int afp_addicon(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);
int afp_geticoninfo(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                    size_t *rbuflen);
int afp_geticon(AFPObj *obj, char *ibuf, size_t ibuflen, char *rbuf,
                size_t *rbuflen);

#endif
