/*!
 * @file
 * Header file for lantest directory cache statistics functions
 */

#ifndef LANTEST_DIRCACHE_STATS_H
#define LANTEST_DIRCACHE_STATS_H

/*! Function declaration for displaying dircache statistics */
void display_dircache_statistics(void);

#endif /* LANTEST_DIRCACHE_STATS_H */